---
author: Dan Gunter
date: 2018-08-21T17:53:00
---
# Building IDAES Docker containers

This document is a guide for developers to building and deploying the IDAES code in Docker.

## General info

Everything for building the containers should be located under the `docker` directory, where you found this document.
The Dockerfiles will be at the top-level of the directory, named `Dockerfile-<tag>`. Scripts to create containers  (and all else Docker-related) will be in a subdirectory called `scripts/`. Overall the layout looks like this:

    idaes : repository directory
      |
      +- docker : this directory
           |
           o- Dockerfile-* : Docker build files
           o- docker-build.sh : main script to run
           +- scripts : scripts to build containers
           |    |
           |    +- shared : shared scripts
           +- shifter : scripts for Shifter/NERSC

To create a container with a given tag, the convention (should be) to run, while in this directory, `./docker-build.sh <tag>`. Some invocations may take additional arguments. This build script will do some output redirection and start the build.

Most of the output will be in a log file, but progress markers from Docker and the scripts themselves will be printed to standard output. A typical run will look like this:

    ---> Building Dockerfile-convergence. Logs in \
         /tmp/docker-convergence-build.log
    Step 1/23 : FROM python:3.7-slim
    ---> af33bb8848e2
    Step 2/23 : MAINTAINER Project IDAES <dkgunter@lbl.gov>
    ---> Using cache
    ---> 9625bb44e890
    Step 3/23 : USER root
    ---> Using cache
    ---> 2133614a6cae
    Step 4/23 : ADD scripts /usr/local/src/bootstrap
    ---> Using cache
    ---> 89f796767b53
    Step 5/23 : WORKDIR /usr/local/src/bootstrap
    ---> Using cache
    ---> b15e35b007b3
    Step 6/23 : RUN bash shared/install_packages.sh
    ---> Running in e772fd3bca4b
    ---> Upgrade APT

    ....etc...


## Current tags / containers

* convergence: For running (parallel) convergence tests. The command to run is
under `scripts/convergence/run.sh`.

## Running docker

Once you have built the container, you need to use Docker to run an instance of it. This is only a brief summary. Refer to [docs.docker.com](https://docs.docker.com/) for details on how to use Docker.

### Locally

To run locally using Docker on your laptop or desktop, use the "docker run" command, along with the tag for the container, e.g. `docker run convergence`.

To inspect a container, you can use the standard invocation for Docker's interactive mode: `docker run -it convergence /bin/bash`. Adding the shell at the end should drop you into a bash shell.

### NERSC and Shifter

See [this page](http://www.nersc.gov/users/software/using-shifter-and-docker/using-shifter-at-nersc) for details on running Shifter at NERSC. You will need to follow the instructions for "Using NERSC's Private Registry" to get yourself a token. Once you have your token, you need to use it first to set up your credentials for the NERSC registry with your local Docker service.

    laptop$ docker login registry.services.nersc.gov.

Then, you will need to use the "-p" option of the `docker-build.sh` script described above to push the image to NERSC. You can simply re-run the script: `./scripts/docker-build.sh -p <TAG>`. This should result in a bunch of output from Docker showing the image layers being pushed over the network to the NERSC registry.

Next you need to login to Shifter at NERSC and "pull" the image. Replace '<TOKEN>' with the token and '<TAG>' with the image tag:

    laptop$ ssh <some NERSC host, e.g. cori.nersc.gov>
    cori$ module load shifter-registry  # load shifter scripts
    cori$ shifterimg login # login
    default username: dang
    default password: <TOKEN>
    cori$ shifterimg -v pull registry.services.nersc.gov/dang/<TAG>:latest

To run on NERSC you will need some more stuff in your SLURM batch file.
Example files are provided in the `shifter/` subdirectory.

## Questions and Comments

Questions and comments on this guide are welcome. Please file an issue in the ["idaes" Github repository](https://github.com/IDAES/idaes) with full details on what was confusing, wrong, or in doubt.
