/*
  Trying to create a physical properties pacakge for simple 
  MEA absorber model.  
  
  Author: John Eslick
  Date: August 1, 2016 
  
  Notes:
  * According to examples hession matrix is initialized to 0 and
    stores uper right triangle (column major).
  * According to my experience the first derivatives are not 
    initialized to 0.
*/
#include <time.h>
#include <math.h>
#include <iostream>
//This may be bad but both ADOL-C and ASL define "function" so I'm 
//changing "function" to "function_old" in adol-c, I don't need it
#define function function_old
#include "adolc/adolc.h"
#undef function
#include <adolc/adalloc.h>
#include "funcadd.h"

static const real mw[] = {  //Component molecular weights in kg/mol
    0.04401,  // CO2
    0.01802,  // H20
    0.02801,  // N2
    0.06108,  // MEA
    0.03200}; // O2

static const real R_gas = 8.314472;
static const double Tref = 298.15;
static const double Pref = 101325;

//Reuse tapes or not.  Mostly for speed testing. As far as I know
//you will always want to reuse the tapes.
#define REUSE_AD_TAPES

//Number of components
static const int N_COMP=5;

typedef enum{ //Component indexes
    COMP_CO2 = 0,
    COMP_H2O = 1,
    COMP_N2 = 2,
    COMP_MEA = 3,
    COMP_O2 = 4
}comp_indx;

typedef enum{ //Tape tags, need to keep track for reuse
    TTAG_NO_SAVE = 0, //Properties where don't want to save the tape
    TTAG_MW_AVE = 1,  //Functions to save with single tape
    TTAG_RHO_VAP = 2,
    TTAG_V_VAP = 3,
    TTAG_CP_VAP = 4,
    TTAG_MU_VAP = 5,
    TTAG_MU_LIQ = 6,
    TTAG_V_LIQ = 7,
    TTAG_DCO2 = 8,
    TTAG_SIGMA = 9,
    TTAG_HENRY = 10,
    TTAG_RHO_LIQ = 11,
    TTAG_CP_LIQ = 12,
    TTAG_HVAP_H2O = 13,
    TTAG_HVAP_MEA = 14,
    TTAG_HABS_CO2 = 15,
    TTAG_H_VAP = 16,
    TTAG_D_VAP = 100, //Functions with mutiple tapes for components
                      //Need to space out enough.
    TTAG_VAP_P = 110,
    TTAG_KEQ_P = 120,
    TTAG_END = 140
}tape_tag;

//A count of the number of times a function is called
static unsigned int prop_count[TTAG_END] = {0};

//Try some caching
static double *x_cache[TTAG_END];
static double *g_cache[TTAG_END];
static double **h_cache[TTAG_END];
static double f_cache[TTAG_END];
static unsigned char d_alloc[TTAG_END] = {0};
static unsigned char f_calc[TTAG_END];
static unsigned char g_calc[TTAG_END];
static unsigned char h_calc[TTAG_END];

//Some stuff for hessian
//Xppp[tag]
static double UppR0[2] = {1, 0};
static double *Upp[1] = {UppR0};
double*** Xppp[TTAG_END];
double*** Yppp[TTAG_END];
double*** Zppp[TTAG_END];

static inline unsigned int h_indx(unsigned int i, unsigned int j){
    /*
     * Return the index in the upper right triangle of Hessian matrix
     * given n variables where i and j are the variable indexes.
     * 
     * */
    return i + (j*j + j)/2;
}

int hessian3(short tag, int n, double* argument, double *y, double *grad, double** hess) {
    //modified version of the adol-c hessian2 driver.
    int rc, i,j;
    rc = hov_wk_forward(tag,1,n,1,2,n,argument,Xppp[tag],y,Yppp[tag]);
    MINDEC(rc,hos_ov_reverse(tag,1,n,1,n,Upp,Zppp[tag]));
    for (i=0; i<n; i++){
        grad[i] = Yppp[tag][0][i][0];
        for (j=0;j<=i;j++) hess[i][j] = Zppp[tag][i][j][1];
    }
    return rc;
}

int cache_alloc(short int tag, unsigned int n, unsigned int off){
    unsigned int i,j;
    d_alloc[tag] = 1;
    f_calc[tag] = 0;
    g_calc[tag] = 0;
    h_calc[tag] = 0;
    
    x_cache[tag] = (double*)malloc((n-off)*sizeof(double));
    g_cache[tag] = (double*)malloc((n-off)*sizeof(double));
    h_cache[tag] = (double**)malloc((n-off)*sizeof(double*));
    for(i=0;i<n-off;i++) h_cache[tag][i]=(double*)malloc((i+1)*sizeof(double));
    
    //Save some time in hessian3 by preallocating and not freeing
    Xppp[tag] = myalloc3(n,n,1);
    Yppp[tag] = myalloc3(1,n,1);
    Zppp[tag] = myalloc3(n,n,2);
    
    //Xppp is just identity matrix
    for (i=0; i<n; i++) {
        for (j=0;j<n;j++)
            Xppp[tag][i][j][0] = 0;
        Xppp[tag][i][i][0] = 1;
    }
    return 0;
}

static int derivatives(arglist *al, short int tag, int m, double *r, int off){
    /*
     * Take the ASL arglist and the ADOL-C tape and fill in the 
     * derivative information. Should be same thing for all these 
     * functions, all have scalar return value
     * 
     * Args
     * ----
     * al: ampl argument list
     * tag: tape tag
     * m: size of result vector, 0 to not calculate
     * r: function evaluation results arry (NULL if m=0)
     * 
     * */
    int n = al->n; //number of variables
    static int i,j; //indexes
    static unsigned char same;
    static double one = 1.0;
    
    if(d_alloc[tag]==0){ //Allocate space if not done already
        same = 0;
        cache_alloc(tag, n, off);
    }
    else{ //if space allocated are aruguments same as last time
        same = 1;
        for(i=off;i<n;++i){
            if((double)(al->ra[al->at[i]]) != x_cache[tag][i-off]){
                same = 0;
                break;
            }
        }
    }
    
    //same = 0; // this line should be commented out except for testing
    
    if(!same) for(i=off;i<n;++i){ //Put real args into vector
        // new function args in cache values and indicate that the
        // function value, gradiaent, and hessian have not been 
        // calculated yet
        x_cache[tag][i-off] = (double)(al->ra[al->at[i]]);
        f_calc[tag] = 0;
        g_calc[tag] = 0;
        h_calc[tag] = 0;
    }
    
    //Do calculations if not same as last call
    if(!f_calc[tag] && !al->derivs && !al->hes){
        //Just need function calculation and different from last
        zos_forward(tag,1,n-off,0,x_cache[tag],&f_cache[tag]);
        f_calc[tag] = 1;
    }
    else if(!g_calc[tag] && al->derivs && !al->hes){
        //Need gradient but not hessian, also calculates function  
        zos_forward(tag,1,n-off,1,x_cache[tag],&f_cache[tag]);
        fos_reverse(tag,1,n-off,&one,g_cache[tag]);
        g_calc[tag] = 1;
        f_calc[tag] = 1;
    }
    else if(!h_calc[tag] && al->derivs && al->hes){
        //Need hessian, includes function and gradient
        hessian3(tag, n-off, x_cache[tag], &f_cache[tag], g_cache[tag], h_cache[tag]);
        g_calc[tag] = 1;
        f_calc[tag] = 1;
        h_calc[tag] = 1;
    }
    
    //Copy function value to return value, if wanted
    if(m && r) *r = f_cache[tag];
    
    //Copy hessian and gradient into ASL struct.
    if(al->derivs){
        for(i=0;i<off;++i) al->derivs[i] = 0.0;
        for(i=off;i<n;++i) al->derivs[i] = (real)g_cache[tag][i-off];
        if(al->hes){
            for(i=0;i<n-off;++i){
                for(j=0;j<=i;++j){
                    //ADOL-C hessian is lower left, ASL is upper right
                    //and h[i][j] = h[j][i], ASL initializes elements
                    //of hessian to 0, so don't need to fill 0 in offset
                    al->hes[h_indx(j+off,i+off)] = (real)h_cache[tag][i][j];
                }//end for j
            }//end for i
        }//end if hes
    }//end if derivs
    return 0;
}

static inline char reuse(arglist *al, short int tag, int m, double *r, int off){
    #ifndef REUSE_AD_TAPES
    return 0;
    #else
    if(tag==0) return 0;
    //++prop_count[tag];
    if(prop_count[tag] == 0){ 
        prop_count[tag] = 1;
        return 0;}
    else{
        derivatives(al, tag, m, r, off);
        return 1;}
    #endif
}

static void V_vap_sub(adouble *T, adouble *P, adouble *V){
    *V = R_gas*(*T)/(*P);
}

extern "C" real V_vap(arglist *al){
    /*
     *  Ideal gas molar volume, and AD example
     * 
     * */
    static short int tag=TTAG_V_VAP; //ADOL-C tape tag
    static real V;            //Molar volume
    if(reuse(al, tag, 1, &V, 0)) return V;
    adouble T, P, aV; //Active variables for AD
    
    trace_on(tag); //Start tape for AD
    T <<= al->ra[al->at[0]]; //independent
    P <<= al->ra[al->at[1]]; //independent
    V_vap_sub(&T, &P, &aV);
    aV >>= V; //dependent
    trace_off(); //Stop tape for AD
    
    derivatives(al, tag, 0, NULL, 0);
    return V;
}

extern "C" real cp_ig(arglist *al){
    const int n=al->n;
    static short int tag=TTAG_CP_VAP;
    static real val;
    if(reuse(al, tag, 1, &val, 0)) return val;
    static const real c[N_COMP][4] = {
        {22.243,  5.977e-2, -3.499e-5,  7.464e-9}, // CO2
        {32.218,  0.193e-2,  1.055e-5, -3.593e-9}, // H2O
        {28.883, -0.157e-2,  0.808e-5, -2.871e-9}, // N2
        {24.619,  3.750e-2, -0.138e-5,  0.000}, //These are NH3 in for MEA, until I get real ones
        {25.460,  1.519e-2, -0.715e-5,  1.311e-9}}; //O2
    int i;
    adouble T, acp;
    //Used new here due to lack of VLA support in Visual C++
    adouble *y = new adouble[n-1]; 
    
    trace_on(tag);
    T <<= al->ra[al->at[0]];
    for(i=1;i<n;++i) y[i-1] <<= al->ra[al->at[i]];
    acp = 0;
    for(i=0;i<n-1;++i)
        acp += (c[i][0] + c[i][1]*T + c[i][2]*T*T + c[i][3]*T*T*T)*y[i];
    acp >>= val;
    trace_off();
    derivatives(al, tag, 0, NULL, 0);
    delete[] y;
    return val;
}

extern "C" real h_vap(arglist *al){
    const int n=al->n;
    static short int tag=TTAG_H_VAP;
    static real val;
    if(reuse(al, tag, 1, &val, 0)) return val;
    static const real c[N_COMP][4] = {
        {22.243,  5.977e-2, -3.499e-5,  7.464e-9}, // CO2
        {32.218,  0.193e-2,  1.055e-5, -3.593e-9}, // H2O
        {28.883, -0.157e-2,  0.808e-5, -2.871e-9}, // N2
        {24.619,  3.750e-2, -0.138e-5,  0.000}, //These are NH3 in for MEA, until I get real ones
        {25.460,  1.519e-2, -0.715e-5,  1.311e-9}}; //O2
    int i;
    adouble T, P, h;
    //Used new here due to lack of VLA support in Visual C++
    adouble *y = new adouble[n-2];
    
    trace_on(tag);
    T <<= al->ra[al->at[0]];
    P <<= al->ra[al->at[1]];
    for(i=2;i<n;++i) y[i-2] <<= al->ra[al->at[i]];
    h = 0;
    for(i=0;i<n-2;++i)
        h += (c[i][0]*T + c[i][1]*T*T/2.0 + c[i][2]*T*T*T/3.0 + c[i][3]*T*T*T*T/4.0 
          - (c[i][0]*Tref + c[i][1]*Tref*Tref/2.0 + c[i][2]*Tref*Tref*Tref/3.0 + c[i][3]*Tref*Tref*Tref*Tref/4.0))*y[i];
    h >>= val;
    trace_off();
    derivatives(al, tag, 0, NULL, 0);
    delete[] y;
    return val;
}

extern "C" real rho_vap(arglist *al){
    const int n=al->n;
    static short int tag=TTAG_RHO_VAP;
    static real r;
    if(reuse(al, tag, 1, &r, 0)) return r;
    int i; 
    adouble T, P, ar;
    //Used new here due to lack of VLA support in Visual C++
    adouble *y = new adouble[n-2]; 
    
    trace_on(tag);
    T <<= al->ra[al->at[0]];
    P <<= al->ra[al->at[1]];
    for(i=2;i<n;++i) y[i-2] <<= al->ra[al->at[i]];
    ar = 0;
    for(i=0;i<n-2;++i) ar += mw[i]*y[i];
    ar *= P/(R_gas*T);
    ar >>= r;
    trace_off();
    derivatives(al, tag, 0, NULL, 0);
    delete[] y;
    return r;
}

extern "C" real D_vap(arglist *al){
    /*
     * Return the component vapor diffusivities.
     * 
     * args
     * 0 = Component index (should be a constant)
     * 1 = Temperature
     * 2 = presure
     * 3 to n-1 = y
     * 
     */
    const int n=al->n;
    short int ci = (short int)(al->ra[al->at[0]] + 0.5);
    short int tag=TTAG_D_VAP + ci;
    static real D;
    if(reuse(al, tag, 1, &D, 1)) return D;
    int i; //number of variables
    static const real v[N_COMP] = {26.7, 13.1, 18.5, 58.62, 16.3};
    adouble T, P, aD;
    //Used new here due to lack of VLA support in Visual C++
    adouble *y = new adouble[n-3];
    adouble *bd = new adouble[n-3];
    
    trace_on(tag);
    T <<= al->ra[al->at[1]];
    P <<= al->ra[al->at[2]];
    for(i=3;i<n;++i) y[i-3] <<= al->ra[al->at[i]];
    aD = 0;
    for(i=0;i<n-3;++i){
        bd[i] = 0.010245606028332951*pow(T, 1.75)*sqrt(0.001/mw[ci] + 0.001/mw[i])/
                (P*pow(pow(v[ci],1.0/3.0)+pow(v[i],1.0/3.0),2));
    }
    for(i=0;i<n-3;++i){
        aD += y[i]*bd[i];
    }
    aD >>= D;
    trace_off();
    derivatives(al, tag, 0, NULL, 1);
    delete[] y;
    delete[] bd;
    return D;
}

extern "C" real mw_ave(arglist *al){
    /*
     * Return the average molecular weight of a mixture.
     * 
     * args: component mole fractions (order see comp_indx type)
     * return: ave molecular weight kg/mol
     * 
     */
    const int n=al->n;
    static short int tag=TTAG_MW_AVE;
    static real mwa;
    if(reuse(al, tag, 1, &mwa, 0)) return mwa;
    int i;
    adouble m;
    //Used new here due to lack of VLA support in Visual C++
    adouble *y = new adouble[n]; 
    
    trace_on(tag);
    for(i=0;i<n;++i) y[i] <<= al->ra[al->at[i]];
    m = 0;
    for(i=0;i<n;++i) m += y[i]*mw[i];
    m >>= mwa;
    trace_off();
    derivatives(al, tag, 0, NULL, 0);
    delete[] y;
    return mwa;
}

extern "C" real mu_liq(arglist *al){
    static short int tag=TTAG_MU_LIQ;
    static double val;
    if(reuse(al, tag, 1, &val, 0)) return val;
    static const real a = -0.0838;
    static const real b = 2.8817;
    static const real c = 33.651;
    static const real d = 1817;
    static const real e = 0.00847;
    static const real f = 0.0103;
    static const real g = -2.3890;
    //r: wt fraction MEA
    //mu: viscosity of solvent (aq. MEA)
    adouble T, yCO2, yMEA, r, mu;
   
    trace_on(tag); //AD tape start
    //Independent variables
    T <<= al->ra[al->at[0]];
    yCO2 <<= al->ra[al->at[1]];
    yMEA <<= al->ra[al->at[2]];
    //Equations 
    r = 100*yMEA*mw[3]/(yCO2*mw[0] + yMEA*mw[3] + (1-yMEA-yCO2)*mw[1]);
    mu =  1.002e-3*pow(10,
              (1.3272*(293.15 - T - 0.001053*(T-293.15)*(T-293.15)))/(T-168.15))*
             exp(r*(T*(a*r + b) + c*r + d)*
                 (yCO2/yMEA*(e*r + f*T + g) + 1)/(T*T));
    //Dependent Variables
    mu >>= val;
    trace_off(); //AD tape stop
   
    derivatives(al, tag, 0, NULL, 0);
    return val;
}

extern "C" real mu_vap(arglist *al){
    const int n=al->n;
    static short int tag=TTAG_MU_VAP;
    static real val;
    if(reuse(al, tag, 1, &val, 0)) return val;
    static const real mu0[N_COMP] = 
        {0.01480e-3, 0, 0.01781e-3, 0.01781e-3, 0.02018e-3};
    static const real T0[N_COMP] = {293.15, 0, 300.55, 300.55, 292.25};
    static const real C[N_COMP] =  {240, 0, 111, 111, 127};
    int i, j;
    adouble T, amu=0;
    adouble *y = new adouble[n-1]; 
    adouble *mu = new adouble[n-1]; 
    adouble *theta = new adouble[n-1]; 

    static const double mult2sqrt2 = 2*sqrt(2);
    
    trace_on(tag);
    T <<= al->ra[al->at[0]];
    for(i=1;i<n;++i) y[i-1] <<= al->ra[al->at[i]];
    amu = 0;
    for(i=0;i<n-1; ++i) theta[i] = 0;
    
    for(i=0;i<n-1;++i){ //calculate component viscosities.
        if(i==(int)COMP_H2O){ //water has different formula
            mu[i] = 1.7096e-8*pow(T,1.1146);
        }
        else{
            mu[i] = mu0[i]*(T0[i] + C[i])/(T + C[i])*pow(T/T0[i], 1.5);
        }
    }
    
    for(i=0;i<n-1; ++i){ //calculate theta
        for(j=0;j<n-1;++j)if(i!=j){
            theta[i] += y[j]*pow(1+pow(mu[i]/mu[j], 0.5)*pow(mw[j]/mw[i],0.25),2)/
                (mult2sqrt2*pow(1+mw[i]/mw[j],0.5));
        }
    }
    for(i=0;i<n-1; ++i){
        amu += mu[i]*y[i]/(y[i] + theta[i]);
    }
    amu >>= val;
    trace_off();
    derivatives(al, tag, 0, NULL, 0);
    delete[] y;
    delete[] mu;
    delete[] theta;
    return val;
}

extern "C" real cp_liq(arglist *al){
    /* Calculate the heat capacity of the solvent.  This is only
     * based on Water and MEA
     * 
     * water = 0 and mea = 1 (not using the regular component list)
     * */
    static short int tag=TTAG_CP_LIQ;
    static real cp;
    if(reuse(al, tag, 1, &cp, 0)) return cp;
    static const real a[2][5] = 
              {{2.6161,  3.706e-3, 3.787e-6,         0,          0},
               {4.2107, -1.696e-3, 2.568e-5, -1.095e-7,  3.038e-10}};
    static const int mwmap[2] = {1, 3};  //map component indexs 
    int i;
    adouble y[2], T, acp, cp_pure[2];
    
    trace_on(tag);
    //Set independent variables.
    T <<= al->ra[al->at[0]];
    y[0] <<= al->ra[al->at[1]];
    y[1] <<= al->ra[al->at[2]];
    //Calculte pure component cps
    acp = 0;
    for(i=0;i<2;++i){
        cp_pure[i] = 1000.0*mw[mwmap[i]]*(a[i][0] + a[i][1]*T + 
            a[i][2]*T*T + a[i][3]*T*T*T + a[i][4]*T*T*T*T);
       acp += cp_pure[i]*y[i]; 
    }
    acp >>= cp;
    trace_off();
    
    derivatives(al, tag, 0, NULL, 0);
    return cp;
}

extern "C" real hvap_H2O(arglist *al){
    static short int tag = TTAG_HVAP_H2O;
    static real val;
    if(reuse(al, tag, 1, &val, 0)) return val;
    static const real c[4] = {56600, 0.61204, -0.6267, 0.3988};
    static const real Tc = 647.13; 
    adouble T, h, Tr;
    
    trace_on(tag);
    T <<= al->ra[al->at[0]];
    Tr = T/Tc;
    h = c[0]*pow(1 - Tr, c[1] + c[2]*Tr + c[3]*Tr*Tr);
    h >>= val;
    trace_off();
    
    derivatives(al, tag, 0, NULL, 0);
    return val;
}

extern "C" real hvap_MEA(arglist *al){
    static short int tag = TTAG_HVAP_MEA;
    static real val;
    if(reuse(al, tag, 1, &val, 0)) return val;
    adouble T, h;

    trace_on(tag);
    T <<= al->ra[al->at[0]];
    h = 58000;
    h >>= val;
    trace_off();
    
    derivatives(al, tag, 0, NULL, 0);
    return val;
}

extern "C" real habs_CO2(arglist *al){
    static short int tag = TTAG_HABS_CO2;
    static real val;
    if(reuse(al, tag, 1, &val, 0)) return val;
    adouble T, h;

    trace_on(tag);
    T <<= al->ra[al->at[0]];
    h = 89740 + 126.89*(T - 40);
    h >>= val;
    trace_off();
    
    derivatives(al, tag, 0, NULL, 0);
    return val;
}

extern "C" real V_liq(arglist *al){
    static short int tag = TTAG_V_LIQ;
    static real val;
    if(reuse(al, tag, 1, &val, 0)) return val;
    static const real Vst = -1.8218e-6;
    static const real a[5] = {0.0, -3.2484e-3, 0.0, -5.35162e-4, 0.0};
    static const real b[5] = {0.0, 1.65, 0.0, -0.451417, 0.0};
    static const real c[5] = {927.0, 793.0, 1.0, 1194.51, 1.0};
    int i, n=al->n;
    adouble T, V;
    adouble *y = new adouble[n-1]; 
    
    trace_on(tag);
    T <<= al->ra[al->at[0]];
    for(i=1;i<n;++i) y[i-1] <<= al->ra[al->at[i]];
    V = 0;
    for(i=0;i<n-1;++i) V += y[i]*mw[i]/(a[i]*T*T + b[i]*T + c[i]);
    V += y[COMP_MEA]*y[COMP_H2O]*Vst;
    V >>= val;
    trace_off();
    
    derivatives(al, tag, 0, NULL, 0);
    delete[] y;
    return val;
}

extern "C" real rho_liq(arglist *al){
    const int n=al->n;
    static short int tag = TTAG_RHO_LIQ;
    static real val;
    if(reuse(al, tag, 1, &val, 0)) return val;
    static const real Vst = -1.8218e-6;
    static const real a[5] = {0.0, -3.2484e-3, 0.0, -5.35162e-4, 0.0};
    static const real b[5] = {0.0, 1.65, 0.0, -0.451417, 0.0};
    static const real c[5] = {927.0, 793.0, 1.0, 1194.51, 1.0};
    int i;
    adouble T, V, rho;
    adouble *y = new adouble[n-1];
    
    trace_on(tag);
    T <<= al->ra[al->at[0]];
    for(i=1;i<n;++i) y[i-1] <<= al->ra[al->at[i]];
    V = 0;
    for(i=0;i<n-1;++i) V += y[i]*mw[i]/(a[i]*T*T + b[i]*T + c[i]);
    V += y[COMP_MEA]*y[COMP_H2O]*Vst;
    rho = 0;
    for(i=0;i<n-1;++i) rho += y[i]*mw[i];
    rho /= V;
    rho >>= val;
    trace_off();
    
    derivatives(al, tag, 0, NULL, 0);
    delete[] y;
    return val;
}

extern "C" real D_CO2(arglist *al){
    const int n=al->n;
    static short int tag = TTAG_DCO2;
    static real val;
    if(reuse(al, tag, 1, &val, 0)) return val;
    static const real Vst = -1.8218e-6;
    static const real a[5] = {0.0, -3.2484e-3, 0.0, -5.35162e-4, 0.0};
    static const real b[5] = {0.0, 1.65, 0.0, -0.451417, 0.0};
    static const real c[5] = {927.0, 793.0, 1.0, 1194.51, 1.0};
    static const real d[5] = {5.07e-6,  8.65e-10,  2.78e-13, -2371, -9.34e-2};
    int i;
    adouble T, V, D, C;
    adouble *y = new adouble[n-1]; 
    
    trace_on(tag);
    T <<= al->ra[al->at[0]];
    for(i=1;i<n;++i) y[i-1] <<= al->ra[al->at[i]];
    V = 0;
    for(i=0;i<n-1;++i) V += y[i]*mw[i]/(a[i]*T*T + b[i]*T + c[i]);
    V += y[COMP_MEA]*y[COMP_H2O]*Vst;
    C = y[COMP_MEA]/V;
    D = 0.46351*(d[0] + d[1]*C + d[2]*C*C)*exp((d[3] + d[4]*C)/T)*exp(252/T);
    D >>= val;
    trace_off();
    
    derivatives(al, tag, 0, NULL, 0);
    delete[] y;
    return val;
}

extern "C" real D_MEA(arglist *al){
    static short int tag = TTAG_V_LIQ;
    static real val;
    if(reuse(al, tag, 1, &val, 0)) return val;
    static const real Vst = -1.8218e-6;
    static const real a[5] = {0.0, -3.2484e-3, 0.0, -5.35162e-4, 0.0};
    static const real b[5] = {0.0, 1.65, 0.0, -0.451417, 0.0};
    static const real c[5] = {927.0, 793.0, 1.0, 1194.51, 1.0};
    int i, n=al->n;
    adouble T, C, D;
    adouble *y = new adouble[n-1]; 
    
    trace_on(tag);
    T <<= al->ra[al->at[0]];
    for(i=1;i<n;++i) y[i-1] <<= al->ra[al->at[i]];
    C = 0;
    for(i=0;i<n-1;++i) C += y[i]*mw[i]/(a[i]*T*T + b[i]*T + c[i]);
    C += y[COMP_MEA]*y[COMP_H2O]*Vst;
    C = y[COMP_MEA]/C;
    D = exp(-13.275 - 2198.3/T - 7.8142e-5*C);
    D >>= val;
    trace_off();
    
    derivatives(al, tag, 0, NULL, 0);
    delete[] y;
    return val;
}

extern "C" real D_MEACOO(arglist *al){
    static short int tag=TTAG_MU_LIQ;
    static double val;
    if(reuse(al, tag, 1, &val, 0)) return val;
    static const real a = -0.0838;
    static const real b = 2.8817;
    static const real c = 33.651;
    static const real d = 1817;
    static const real e = 0.00847;
    static const real f = 0.0103;
    static const real g = -2.3890;
    //r: wt fraction MEA
    //mu: viscosity of solvent (aq. MEA)
    adouble T, yCO2, yMEA, r, mu, D;
   
    trace_on(tag); //AD tape start
    //Independent variables
    T <<= al->ra[al->at[0]];
    yCO2 <<= al->ra[al->at[1]];
    yMEA <<= al->ra[al->at[2]];
    //Equations 
    r = 100*yMEA*mw[3]/(yCO2*mw[0] + yMEA*mw[3] + (1-yMEA-yCO2)*mw[1]);
    mu =  1.002e-3*pow(10,
              (1.3272*(293.15 - T - 0.001053*(T-293.15)*(T-293.15)))/(T-168.15))*
             exp(r*(T*(a*r + b) + c*r + d)*
                 (yCO2/yMEA*(e*r + f*T + g) + 1)/(T*T));
    //Dependent Variables
    D = exp(-22.64 - 1000.0/T - 0.7*log(mu));
    D >>= val;
    trace_off(); //AD tape stop
   
    derivatives(al, tag, 0, NULL, 0);
    return val;
}

extern "C" real surface_tension(arglist *al){
    //The arguments are T, yCO2, yH2O, yMEA
    static short int tag=TTAG_SIGMA;
    static real sigma; //return value mixture surface tension
    if(reuse(al, tag, 1, &sigma, 0)) return sigma;
    static const int cmap[3] = {0,1,3}; //this clist to main clist
    static const int iCO2 = 0;
    static const int iH2O = 1;
    static const int iMEA = 2;
    //Parameters
    static const real Fa =  2.4558;
    static const real Fb = -1.5311;
    static const real Fc =  3.4994;
    static const real Fd = -5.6398;
    static const real Fe = 10.2109;
    static const real Ff =  2.3122;
    static const real Fg =  4.5608;
    static const real Fh = -2.3924;
    static const real Fi =  5.3324;
    static const real Fj = -12.0494;
    static const real Tc[3] = {0.0, 647.13,  614.45}; //don't need CO2
    static const real c[3][6] = {
        {-5.987,   3.7699, -0.43164, 0.018155, -0.01207, 0.002119},
        { 0.18548, 2.717,  -3.554,   2.047,     0,       0},
        { 0.09945, 1.067,   0,       0,         0,       0}};
    // sigi: pure component surface tension
    // r: weight fraction CO2
    // asig: mixture CO2
    // alpha: CO2 loading (mol CO2)/(mol MEA)
    static adouble y[3], T, asig, sigi[3], alpha, r;
    int i;
    
    trace_on(tag); //start the tape for AD
    //Independent Variables
    T <<= al->ra[al->at[0]];
    y[0] <<= al->ra[al->at[1]]; //CO2
    y[1] <<= al->ra[al->at[2]]; //H2O
    y[2] <<= al->ra[al->at[3]]; //MEA
    
    //CO2 weight fraction
    r = y[iCO2]*mw[cmap[iCO2]]/
        (y[iCO2]*mw[cmap[iCO2]] + 
         y[iH2O]*mw[cmap[iH2O]] + 
         y[iMEA]*mw[cmap[iMEA]]);
    //CO2 loading
    alpha = y[iCO2]/y[iMEA];

    //pure h2o and MEA
    for(i=1;i<3;++i){ //{1, 2} = {H2O, MEA}
        sigi[i] = c[i][0]*pow((1-T/Tc[i]), 
                c[i][1] + c[i][2]*T/Tc[i] + c[i][3]*(T/Tc[i])*(T/Tc[i]));
    }
    //pure CO2 (well probably not exactly.)
    sigi[iCO2] =     c[iCO2][0]*r*r + c[iCO2][1]*r + c[iCO2][2] + 
                  T*(c[iCO2][3]*r*r + c[iCO2][4]*r + c[iCO2][5]);
    //Mixture surface tension.
    asig = sigi[iH2O] + 
        (sigi[iH2O] - sigi[iCO2])*y[iCO2]*(Fa + Fb*alpha + Fc*alpha*alpha + Fd*r + Fe*r*r) +
        (sigi[iMEA] - sigi[iH2O])*y[iMEA]*(Ff + Fg*alpha + Fh*alpha*alpha + Fi*r + Fj*r*r);
    //dependent variable
    asig >>= sigma;
    trace_off(); //stop the tape for AD
    derivatives(al, tag, 0, NULL, 0); //Calculate derivatives.
    return sigma;
}

extern "C" real henry_co2(arglist *al){
    static short int tag=TTAG_HENRY;
    static real val; //CO2 henry's constant return value
    if(reuse(al, tag, 1, &val, 0)) return val;
    //Constants
    static const real a_N2O_H2O[4] = {158.245, -9048.596, -20.860, -0.00252};
    static const real a_CO2_H2O[4] = {145.369, -8172.355, -19.303,  0.0};
    static const real a_N2O_MEA[2] = {-9172.50, 39.598};
    //Variables
    adouble T, yH2O, yMEA, H_CO2, H_CO2_H2O, H_N2O, H_N2O_MEA, H_N2O_H2O;

    trace_on(tag); //start AD tape
    //Independent variables
    T <<= al->ra[al->at[0]];
    yH2O <<= al->ra[al->at[1]];
    yMEA <<= al->ra[al->at[2]];
    //Equations
    H_CO2_H2O = exp(a_CO2_H2O[0] + 
                    a_CO2_H2O[1]/T + 
                    a_CO2_H2O[2]*log(T) + 
                    a_CO2_H2O[3]*T);
    H_N2O_H2O = exp(a_N2O_H2O[0] + 
                    a_N2O_H2O[1]/T + 
                    a_N2O_H2O[2]*log(T) + 
                    a_N2O_H2O[3]*T);
    H_N2O_MEA = a_N2O_MEA[0] + a_N2O_MEA[1]*T;
    H_N2O = yH2O*H_N2O_H2O + yMEA*H_N2O_MEA + 
             3.524641533*(yMEA*yH2O)*(yMEA*yH2O)*
             (1-T/324.718)*exp(-13.219*yMEA);
    H_CO2 = H_CO2_H2O*H_N2O/H_N2O_H2O;
    //Dependent variable
    H_CO2 >>= val;
    trace_off(); //stop AD tape
    // Get derivatives.
    derivatives(al, tag, 0, NULL, 0);
    return val;
}

extern "C" real vapor_pressure(arglist *al){
    short int ci = (short int)(al->ra[al->at[0]] + 0.5);
    short int tag = TTAG_VAP_P + ci;
    static real val;
    if(reuse(al, tag, 1, &val, 1)) return val;
    static const real a[5] = {0.0, 72.55, 0.0, 172.78};
    static const real b[5] = {0.0, -7206.70, 0.0, -13492,};
    static const real c[5] = {0.0, -7.1385, 0.0, -21.914};
    static const real d[5] = {0.0, 4.05e-6, 0.0, 1.38e-5};
    adouble y, T, P;

    trace_on(tag);
    T <<= al->ra[al->at[1]];
    y <<= al->ra[al->at[2]];
    P = y*exp(a[ci] + b[ci]/T + c[ci]*log(T) + d[ci]*T*T);
    P >>= val;
    trace_off();
    derivatives(al, tag, 0, NULL, 1); //Calculate derivatives.
    return val;
}

extern "C" real logKeq(arglist *al){
    short int ci = (short int)(al->ra[al->at[0]] + 0.5);
    short int tag=TTAG_KEQ_P + ci;
    static real val;  
    if(reuse(al, tag, 1, &val, 1)) return val;
    static const real a[4][10] = {
        //0,      1,       2,        3,     4, 5, 6, 7,     8,       9
        { 0,  -13445.9, -12092.1, -12431.7, 0, 0, 0, 0, -3090.83, -5851.11},
        { 0,  -22.4773, -36.7816, -35.4819, 0, 0, 0, 0,  0.0,      0.0},
        { 0,   140.932,  235.482,  220.067, 0, 0, 0, 0,  6.69425, -3.3636},
        { 0,   log(1e6), log(1e3), log(1e3),0, 0, 0, 0,  log(1e3), log(1e3)}};
    adouble T, K;
    trace_on(tag);
    T <<= al->ra[al->at[1]];
    K = a[0][ci]/T + a[1][ci]*log(T) + a[2][ci] + a[3][ci];
    K >>= val;
    trace_off();
    derivatives(al, tag, 0, NULL, 1); //Calculate derivatives.
    return val;
}

extern "C" real Keq(arglist *al){
    short int ci = (short int)(al->ra[al->at[0]] + 0.5);
    short int tag=TTAG_KEQ_P + ci;
    static real val;  
    if(reuse(al, tag, 1, &val, 1)) return val;
    static const real a[4][10] = {
        //0,      1,       2,        3,     4, 5, 6, 7,     8,       9
        { 0,  -13445.9, -12092.1, -12431.7, 0, 0, 0, 0, -3090.83, -5851.11},
        { 0,  -22.4773, -36.7816, -35.4819, 0, 0, 0, 0,  0.0,      0.0},
        { 0,   140.932,  235.482,  220.067, 0, 0, 0, 0,  6.69425, -3.3636},
        { 0,   log(1e6), log(1e3), log(1e3),0, 0, 0, 0,  log(1e3), log(1e3)}};
    adouble T, K;
    trace_on(tag);
    T <<= al->ra[al->at[1]];
    K = exp(a[0][ci]/T + a[1][ci]*log(T) + a[2][ci] + a[3][ci]);
    K >>= val;
    trace_off();
    derivatives(al, tag, 0, NULL, 1); //Calculate derivatives.
    return val;
}

void funcadd(AmplExports *ae){
    /* Arguments for addfunc (this is not fully detailed see funcadd.h)
     * 1) Name of function in AMPL
     * 2) Function pointer to C function
     * 3) see FUNCADD_TYPE enum in funcadd.h
     * 4) Number of arguments (the -1 is variable arg list length)
     * 5) Void pointer to function info
     */
    //Real value function, and suppress anoying warings that, I think 
    //happen when this gets called on an already loaded library.
    int t = 0;
    addfunc("mw_ave", (rfunc)mw_ave, t, -1, NULL);
    addfunc("cp_ig", (rfunc)cp_ig, t, -1, NULL);
    addfunc("cp_liq", (rfunc)cp_liq, t, 3, NULL);
    addfunc("V_vap", (rfunc)V_vap, t, 2, NULL);
    addfunc("V_liq", (rfunc)V_liq, t, -1, NULL);
    addfunc("rho_vap", (rfunc)rho_vap, t, -1, NULL);
    addfunc("rho_liq", (rfunc)rho_liq, t, -1, NULL);
    addfunc("mu_liq", (rfunc)mu_liq, t, -1, NULL);
    addfunc("mu_vap", (rfunc)mu_vap, t, -1, NULL);
    addfunc("D_vap", (rfunc)D_vap, t, -1, NULL);
    addfunc("surface_tension", (rfunc)surface_tension, t, -1, NULL);
    addfunc("vapor_pressure", (rfunc)vapor_pressure, t, 3, NULL);
    addfunc("henry_co2", (rfunc)henry_co2, t, -1, NULL);
    addfunc("hvap_H2O", (rfunc)hvap_H2O, t, 1, NULL);
    addfunc("hvap_MEA", (rfunc)hvap_MEA, t, 1, NULL);
    addfunc("habs_CO2", (rfunc)habs_CO2, t, 1, NULL);
    addfunc("logKeq", (rfunc)logKeq, t, 2, NULL);
    addfunc("Keq", (rfunc)Keq, t, 2, NULL);
    addfunc("D_CO2", (rfunc)D_CO2, t, -1, NULL);
    addfunc("D_MEA", (rfunc)D_MEA, t, -1, NULL);
    addfunc("D_MEACOO", (rfunc)D_MEACOO, t, -1, NULL);
    addfunc("h_vap", (rfunc)h_vap, t, -1, NULL);
}
