##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Upload properties file, in JSON form, to Google cloud datastore.
"""
# System imports
import argparse
import json
import logging
import os
import re
import sys
# Local imports
from idaes.dmf.propdb.util import get_logger

_log = get_logger()




class Column(object):
    """Generic, abstract column
    """
    type_name = 'generic'
    def __init__(self, name, data):
        self.name = name
        self.units = data['units']
        self.values = data['values']
        self.entity_type = self.type_name


class PropertyColumn(Column):
    """Data column for a property.
    """
    type_name = 'Property'
    def __init__(self, name, data):
        Column.__init__(self, name, data)
        self.errors = data['errors']
        self.error_type = data['error_type']

    def data(self):
        return {
            'units': self.units,
            'values': self.values,
            'errors': self.errors,
            'error_type': self.error_type
        }


class StateColumn(Column):
    """Data column for a state.
    """
    type_name = 'State'

    def data(self):
        return {
            'units': self.units,
            'values': self.values
        }


if __name__ == '__main__':
    sys.exit(main())
