##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Example of merging
"""
from six import StringIO

from idaes.dmf import propdata


def main():
    # data
    csv = ('Num,State (units 1),Typeof1 Error,'
           'Prop (units 2),Typeof2 Error\n'
           '1, 1.0, 0, 100.0, 0.1\n'
           '2, 2.0, 0, 200.0, 0.2\n'
           '3, 3.0, 0, 300.0, 0.3\n',
           'Num,State (units 1),Typeof1 Error,Prop (units 2),Typeof2 Error,'
           'PropB (units 2),Typeof1B Error\n'
           '4, 4.0, 0, 400.0, 0.4, -400.0, -0.4\n'
           '5, 5.0, 0, 500.0, 0.5, -500.0, -0.5\n',
           'Num,State (units 1),Typeof1 Error,'
           'PropB (units 2),Typeof1B Error,'
           'PropC (units 2),Typeof2B Error,'
           'PropD (units 2),Typeof3B Error\n'
           '6, 6.0, 0, -600.0, 0.6, 60.0, 0.6, -60.0, -0.6\n'
           '7, 7.0, 0, -700.0, 0.7, 70.0, 0.7, -70.0, -0.7\n')
    data = propdata.PropertyData.from_csv(StringIO(csv[0]), 1)
    print('initial data has {} rows and {} columns'.format(
        data.num_rows, data.num_columns))
    # add more data
    for i in (1, 2):
        data.add_csv(StringIO(csv[i]))
        print('updated data has {} rows and {} columns'.format(
            data.num_rows, data.num_columns))
    # show values for selected columns
    for colname in 'Prop', 'PropB', 'PropD':
        values = data.get_column(colname).values
        print('{} column values: {}'.format(colname, values))

    return 0

if __name__ == '__main__':
    import sys
    sys.exit(main())
